#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cmath>
#include <stdio.h>  
#include <stdlib.h>   

//#define USE_BOOST
#ifdef USE_BOOST
#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#endif

using namespace std;


// globale Variablen
const int N = 20;
const int M = 1001; 


//=========================================================
// I/O
//=========================================================

// write 2D data
void write2D( const double x[M], const double y[M], const char* filename ) {
  ofstream out ( filename );
  for( int i = 0; i < M; i++ )
    out << x[i] << " " << y[i] << endl;
}

// read 2D data
void read2D( double x[M], double y[M], const char* filename ) {
    
  std::ifstream readfile ( filename );
  std::string line;
  
  int i = 0;
  while( (!readfile.eof()) && (i < M) ){
   
   std::getline( readfile, line );
   std::stringstream ss( line );
   ss >> x[i] >> y[i];
   i++;
  }

}

//=========================================================
// CONSOLE OUTPUT
//=========================================================

//
void printMatrix ( double A[M][N] ) {
  for( int i = 0; i < M; i++ ){
    for( int j = 0; j < N; j++ )
      cerr << A[i][j] << " ";
    cerr << endl;
  }
  cerr << endl;
}

//
void printVector ( double x[N] ) {
  for( int j = 0; j < N; j++ )
    cerr << x[j] << " ";
  cerr << endl;
  cerr << endl;
}

//=========================================================
// DATA GENERATION
//=========================================================

// evaluate piecewise defined zig-zag-function
void sampleZigzagData( double t[M], double z[M] ){  
  for( int i = 0; i < M; i++ ){
    t[i] = 2. * M_PI * i / (M-1.);
    if( t[i] < M_PI / 2. ){
      z[i] = t[i];
    }
    else if( t[i] < 3 * M_PI / 2. ){
      z[i] = M_PI - t[i];
    }
    else{
      z[i] = t[i] - 2 * M_PI;
    }
  }
}

#ifdef USE_BOOST
// zigzag function with error
void generateZigzagData( double t[M], double z[M], double eps ){
  //srand (time(NULL));
  
  boost::mt19937 *rng = new boost::mt19937();
  rng->seed(time(NULL));
  boost::normal_distribution<> distribution(0, eps);
  boost::variate_generator< boost::mt19937, boost::normal_distribution<> > dist(*rng, distribution);

  
  for( int i = 0; i < M; i++ ){
    t[i] = 2. * M_PI * i / (M-1.);
    if( t[i] < M_PI / 2. ){
      z[i] = t[i];
    }
    else if( t[i] < 3 * M_PI / 2. ){
      z[i] = M_PI - t[i];
    }
    else{
      z[i] = t[i] - 2 * M_PI;
    }
    //z[i] += (eps * rand() / RAND_MAX) - 0.5 * eps;
    z[i] += dist();
  }
}
#endif

//=========================================================
// SYSTEM SETUP
//=========================================================

//
void buildSystemForSinusInterpolation( double A[M][N] ) {
  if( M != N )
    throw "ERROR in buildSystemForSinusInterpolation()! Number of rows and columns should be equal!";
  
   double max = 2 * M_PI;
   for( int i = 0; i < N; i++ ){
    A[i][0] = 1.;
    for( int j = 1; j < N; j++ )
      A[i][j] = A[i][j-1] * i * max / (N-1);
   }
}

//
void buildSystemForPoissonProblem( double A[M][N] ) {
  
   if( M != N )
     throw "ERROR in buildSystemForPoissonProblem()! Number of rows and columns should be equal!";
    
   for( int i = 0; i < N; i++ ){
    
    // Diagonale und Nebendiagonale
    A[i][i] = 2. * N * N;
    if( i > 0 )
      A[i][i-1] = -1. * N * N;
    if( i < N-1 )
      A[i][i+1] = -1. * N * N;
    
    // Nulleintraege
    for( int j = i+2; j < N; j++ )
      A[i][j] = 0.;
    for( int j = 0; j < i-1; j++ )
      A[i][j] = 0.;
   }
}

//
void buildMatrix( double A[M][N] ) {
  for( int i = 0; i < M; i++ )
    for( int j = 0; j < N; j++ )
      A[i][j] = 0;
    
  A[0][0] = 2.; A[0][2] = 1.;
  A[1][1] = 2.; A[1][2] = 7.;
  A[2][1] = 6.; 
  A[3][1] = 3;
  A[4][0] = std::sqrt( 12 ); A[4][2] = std::sqrt( 3 );
}


//
void buildFourierSystem( double t[M], double A[M][N] ) {
  for( int i = 0; i < M; i++ )
    for( int j = 0; j < N; j++ )
      A[i][j] = sin( (j+1) * t[i] );
}

//=========================================================
// QR - ALGORITHM
//=========================================================

// QR - Zerlegung mit Householder
void computeQR( double A[M][N], double b[M] ) {
     
  if( M < N )
   throw "ERROR in computeQR()! Number of rows should no be smaller than number of columns!";
  
  // N = min( M, N )
  for( int i = 0; i < N; i++ ){
    
    // z_i ist i-te Spalte (unterhalb des Diagonalelements)
    double z[M-i];
    double normSqr = 0.;
    for( int j = 0; j < M-i; j++ ){
      z[j] = A[i+j][i];
      normSqr += z[j] * z[j];
    }
    
    // v_i ist Normale der Ebene, an der gespiegelt wird
    // v_i = z_i + sgn(z(0)) * |z_i| * e_1
    double firstComp = z[0];
    double signZ = (firstComp < 0) ? -1. : 1.;
    double norm = std::sqrt( normSqr );
    z[0] += signZ * norm;
    
    // norm of v_i 
    double normReflec = std::sqrt( 2 * norm * (norm + signZ * firstComp) );
    for( int j = 0; j < M-i; j++ )
      z[j] /= normReflec;
    
    // v_i^T b
    double vTb = 0;
    for( int j = 0; j < M-i; j++ )
      vTb += z[j] * b[i+j];

    // H_i * A, H_i = \id - 2 v_i v_i^T
    for( int j = 0; j < N-i; j++ ){
      double vTAj = 0;
      for( int k = 0; k < M-i; k++ )
	vTAj += z[k] * A[i+k][i+j];
      for( int k = 0; k < M-i; k++ )
         A[i+k][i+j] -= 2. * vTAj * z[k];      
    }
    
    // H_i * b
    for( int j = 0; j < M-i; j++ )
      b[i+j] -= 2. * vTb * z[j];
  }

}

//
double solveSystemWithQR( double A[M][N], double rhs[M], double sol[N] ){  
  
  computeQR( A, rhs );   
    
  // backward
  for( int i = N-1; i != -1; i-- ){
    double sum = 0.;
    for( int j = i+1; j < N; j++ )
      sum += A[i][j] * sol[j];
    sol[i] = (rhs[i] - sum) / A[i][i];
  }
  
  // compute error
  double err = 0.;
  for( int j = N; j < M; j++ )
    err += rhs[j] * rhs[j];
  return std::sqrt( err );
  
}

// z_k = \sum_i x_i sin(i t_k) for k = 1, ..., M and x being the coefficients
void evaluateFourierSeries( double t[M], double z[M], double coeff[N] ) { 
  for( int i = 0; i < M; i++ ){
    z[i] = 0.;
    for( int j = 0; j < N; j++ )
      z[i] += coeff[j] * sin( (j+1) * t[i] );
  }
}


//=========================================================
//=========================================================

// Hauptprogramm
int main () {
    
  try{      
    
  static double A[M][N];
  static double b[M];
  static double sol[N];  
  static double t[M];
  
  // ---------------------------------  
  // AUFGABE 3
  // ---------------------------------
  
  //buildMatrix( A );
  //buildSystemForSinusInterpolation( A );
  //buildSystemForPoissonProblem( A );  
  //printMatrix( A );
  
  // ---------------------------------
  // AUFGABE 4
  // ---------------------------------
  
  // AUFGABE 4(a)
  sampleZigzagData( t, b );
  
  // AUFGABE 4(b)
  //read2D( t, b, "signal.dat" ); 
  
  // write data
  //write2D( t, b, "data1.dat" );
  
  // ---------------------------------
  // ---------------------------------
  
  // build up system matrix
  buildFourierSystem( t, A );

     
  clock_t begin, end;
  
  begin = clock();   
  double error = solveSystemWithQR( A, b, sol );
  end = clock();
  
  cerr << "Time = " << double(end - begin) / CLOCKS_PER_SEC << endl;
  
  cerr << "Error = " << error << endl << endl;
  
  //printMatrix( A );  
  printVector( sol );

  // write solution
  evaluateFourierSeries( t, b, sol );
  write2D( t, b, "data2.dat" );

  
  }
  catch ( const char* exc ){
    cerr << "Exception caught: " << exc << endl;
  }

  return 0;
}